public class AVLTree<T extends Comparable<T>> {

    private class Node {		// ein Knoten im Baum
	T    value;			// Wert am Knoten
        Node left;			// Wurzel linker Unterbaum
	Node right;			// Wurzel rechter Unterbaum
	int  balance;			// hoehe(linker Unterbaum) - hoehe(rechter Unterbaum)

	Node(T val) {
	    value = val;
	    left = null;
	    right = null;
	    balance = 0;
	}
    }

    private Node root;			// Wurzel des gesamten AVL-Baums
    private int size;			// Anzahl Knoten im AVL-Baum

    public AVLTree() {
        root = null;
	size = 0;
    }

    public int size() {
        return size;
    }

    public boolean contains(T elem) {
        Node node = root;
        int  diff;

	while (node != null) {
	    diff = node.value.compareTo(elem);
	    if (diff == 0) {
	        return true;
	    }
	    else if (diff>0) {
	        node = node.left;
	    }
	    else {
	        node = node.right;
	    }
	}

	return false;
    }

    public void sortedArray(T[] elems) {
        int index;

        if (elems.length < size) {  
            throw new IllegalArgumentException();
        }   

        fillArray(root,elems,0);        // schreibe vollstaendigen Baum (Wurzel) nach Feld ab Feldanfang
    }   

    private int fillArray(Node node, T[] elems, int index) {
        if(node!=null) {                                // nur wenn Unterbaum nicht leer
            index = fillArray(node.left,elems,index);   // kopiere linken Unterbaum ins Feld
            elems[index] = node.value;                  // kopiere Wert am Knoten in Feld und
            index++;                                    // kopiere rechten Unterbaum ins Feld.
            index = fillArray(node.right,elems,index);
        }   

        return index;
    }

    public void insert(T elem) {
	insert(root,null,elem);
    }

    /*
     * Rekursives Einfuegen in einen Unterbaum.
     *
     * node: Der Wurzelknoten des Baumes bzw. Unterbaums, in den eingefuegt wird.
     * father: Vaterknoten von node, node ist linker oder rechter Sohn von father
     * elem: Der einzufuegende Wert
     *
     * return: true, falls der Unterbaum durch das Einfuegen hoeher geworden ist,
     *    sonst false
     */
    private boolean insert(Node node, Node father, T elem) {
        int     diff;
	boolean grown;
	int     balanceDelta;

        if (node==null) {			// am leeren Unterbaum angekommen
	    node=new Node(elem);		// dann wird hier ein neuer Knoten erzeugt
	    if (father==null) {			// nur moeglich, wenn der root==null also insgesamt leerer Baum
	        root = node;			// dann besteht der Baum jetzt aus genau einem Knoten
	    }					// ansonsten muessen wir untersuchen, ob der neue Knoten
            else if (father.value.compareTo(elem) > 0) {	// linker oder rechter Sohn seines Vaters wird.
	        father.left = node;		// Wert im Vater ist groesser, dann wird neuer Knoten
	    }					// linker Sohn
	    else {				// Wert im Vater ist kleiner, dann wird neuer Knoten
	        father.right = node;		// rechter Sohn
	    }
            size++;				// Baum hat insgesamt einen Knoten mehr

	    return true;			// Unterbaum mit node als Wurzel ist gewachsen
	}

	// Wenn wir hier angekommen sind, gilt node != null.
	// Wir muessen also in einen der beiden Unterbaeume von node einfuegen

	diff = node.value.compareTo(elem);	
	if (diff == 0) {			// Wert elem schon im Baum vorhanden
	    return false;			// kein Einfuegen!
	}
	else if (diff > 0) {			// node.value > elem ==> in linken Unterbaum einfuegen
	    grown = insert(node.left,node,elem);
	    balanceDelta = 1;
	}
	else {					// node.value < elem ==> in rechten Unterbaum einfuegen
	    grown = insert(node.right,node,elem);
	    balanceDelta = -1;
	}

	if (grown) {				// wenn der unterbaum von node, in den wir eingefuegt haben,
	    node.balance += balanceDelta;	// gewachsen ist, dann aendert sich der balance-Wert
	}					// am Knoten node

	if (!grown || node.balance==0) {	// Keine Hoehenaenderung oder Fall 2 (Folie 260)
	    return false;			// Baum mit Wurzel node ist insgesamt nicht gewachsen
	}

	// Wenn wir hier angekommen sind, ist der Baum mit node als Wurzel gewachsen.
	// Es liegt entweder Fall 1 oder Fall 3 vor.

	if (node.balance==1 || node.balance==-1) {	// AVL Kriterium noch erfuellt
	    return true;				// Fall 1 (Folie 257)
	}						// Der Baum ist aber auf jeden Fall gewachsen

	// Wenn wir hier angekommen sind, ist der Baum mit Wurzel node gewachsen und das
	// AVL-Kriterium ist verletzt.
        if (node.balance==2) {
	    if (node.left.balance==1) {
	        rotateLeftLeft(node, father);			// Fall 3a (Folie 265)
	    }
	    else if (node.left.balance==-1) {
	        rotateLeftRight(node, father);			// Fall 3b (Folie 268)
	    }
	    else {
	        assert false : "Illegaler Zustand im AVL-Baum";
	    }
	}
	else if (node.balance==-2) {
	    if (node.right.balance==-1) {
	        rotateRightRight(node, father);			// Fall symmetrisch zu 3a bzw. Folie 265
	    }
	    else if (node.right.balance==1) {
	        rotateRightLeft(node, father);			// Fall symmterisch zu 3b bzw. Folie 268
	    }
	    else {
	        assert false : "Illegaler Zustand im AVL-Baum";
	    }
	}
	else {
	    assert false : "Illegaler Zustand im AVL-Baum";
	}

        // Ausgleichsrotationen fuehren beim Einfuegen stets dazu, dass der Unterbaum mit
	// node als Wurzel nicht gewachsen ist.

	return false;
    }

    private void rotateLeftLeft(Node w, Node father) {	// vgl. Folie 265
        Node v  = w.left;
	Node a1 = v.left;
	Node a2 = v.right;
	Node b  = w.right;

	if (father == null) {			// wenn Rotation an Wurzel
	    root = v;				// dann haben wir jetzt eine neue Wurzel
	} else if (w == father.left) {		// ansonsten hat father einen neuen linken
	    father.left = v;			// oder rechten Sohn
	} else {
	    father.right = v;
	}

	v.left = a1;				// Verweise umhaengen
	v.right = w;
	w.left = a2;
        w.right = b;

	v.balance = 0;				// Unterbaeume ab v und w sind jetzt auf jeden Fall
	w.balance = 0;				// vollstaendig ausgeglichen
    }

    private void rotateRightRight(Node w, Node father) { // symmetrisch zu Folie 265
        Node v  = w.right;
	Node a1 = w.left;
	Node a2 = v.left;
	Node b  = v.right;

	if (father == null) {			// wenn Rotation an Wurzel
	    root = v;				// dann haben wir jetzt eine neue Wurzel
	} else if (w == father.left) {		// ansonsten hat father einen neuen linken
	    father.left = v;			// oder rechten Sohn
	} else {
	    father.right = v;
	}

	v.left = w;				// Verweise umhaengen
	v.right = b;
	w.left = a1;
        w.right = a2;

	v.balance = 0;				// Unterbaeume ab v und w sind jetzt auf jeden Fall
	w.balance = 0;				// vollstaendig ausgeglichen
    }

    private void rotateLeftRight(Node w, Node father) {	// vgl. Folie 268
        Node v  = w.left;
	Node u  = v.right;
	Node a1 = v.left;
	Node a2 = u.left;
	Node a3 = u.right;
	Node b  = w.right;
	int  ubalance = u.balance;		// alte balance von u sichern, da u sich aendert

	if (father == null) {			// wenn Rotation an Wurzel
	    root = u;				// dann haben wir jetzt eine neue Wurzel
	} else if (w == father.left) {		// ansonsten hat father einen neuen linken
	    father.left = u;			// oder rechten Sohn
	} else {
	    father.right = u;
	}

        u.left = v;
	u.right = w;
	v.left = a1;
	v.right = a2;
	w.left = a3;
	w.right = b;

	switch (ubalance)
	{
	    case 0 :			// vgl. Folie 268, hoehe(A2)=hoehe(A3)=h-1
	        v.balance = 0;
		w.balance = 0;
	        break;
	    case 1 :			// hoehe(A2)=h-1, hoehe(A3)=h-2
	        v.balance = 0;
	        w.balance = -1;
	        break;
	    case -1 :			// hoehe(A2)=h-2, hoehe(A3)=h-1
	        v.balance = 1;
	        w.balance = 0;
	        break;
	}

	u.balance = 0;
    }

    private void rotateRightLeft(Node w, Node father) {	// symmetrisch zu Folie 268
        Node v  = w.right;
	Node u  = v.left;
	Node a1 = w.left;
	Node a2 = u.left;
	Node a3 = u.right;
	Node b  = v.right;
	int  ubalance = u.balance;		// alte balance von u sichern, da u sich aendert

	if (father == null) {			// wenn Rotation an Wurzel
	    root = u;				// dann haben wir jetzt eine neue Wurzel
	} else if (w == father.left) {		// ansonsten hat father einen neuen linken
	    father.left = u;			// oder rechten Sohn
	} else {
	    father.right = u;
	}

        u.left = w;
	u.right = v;
	w.left = a1;
	w.right = a2;
	v.left = a3;
	v.right = b;

	switch (ubalance)
	{
	    case 0 :			// hoehe(A2)=hoehe(A3)=h-1
	        v.balance = 0;
		w.balance = 0;
	        break;
	    case 1 :			// hoehe(A2)=h-1, hoehe(A3)=h-2
	        w.balance = 0;
	        v.balance = -1;
	        break;
	    case -1 :			// hoehe(A2)=h-2, hoehe(A3)=h-1
	        v.balance = 0;
	        w.balance = 1;
	        break;
	}

	u.balance = 0;
    }

    public void printTree() {
        printTree(root, 0);
    }

    private void printTree(Node node, int level) {
        String s;
        if (node==null) {
            return;
        }

        printTree(node.right,level+1);

        s = "";
        for (int i=0 ; i<level ; i++) {
            s+= "   ";
        }
        System.out.println(s+node.value+"["+node.balance+"]");

        printTree(node.left,level+1);
    }
}
